#include "mex.h"
#include <cmath>

/*
 * alg1_leg4.cpp - core algorithm for legpts4.m
 *
 * fast(er) impl. of legpts.m from chebfun // MATLAB
 * just innerest functions and driver
 *
 *
 *  implemented by Lukas Fath, December 2011 - algorithm adapted from [1].

 *  References:
 *  [1] by Nick Hale, April 2009 in the CHEBFUN-package for MATLAB, 
 *      - algorithm adapted from [2]
 *  [2] A. Glaser, X. Liu and V. Rokhlin, "A fast algorithm for the 
 *      calculation of the roots of special functions", SIAM Journal  
 *      on Scientific Computing", 29(4):1420-1438:, 2007.
 *
 * This is a MEX-file for MATLAB.
 */

/* $Revision: 0.0.0.6 $ */

void alg2_Leg(double P,int n,double& root,double& ders)
{
   // % Approximate first root via asymptotic formula
    double nn = n;
    const double PI = 4* atan(1.0);
    int k = n/2; double theta = PI*(4.0*k-1.0)/(4.0*n+2.0);
    double n3= nn*nn*nn; double n4 = n3*nn;

    double x1 = (1-(nn-1)/(8*n3)-1/(384*n4)*(39-28/(sin(theta)*sin(theta))))*cos(theta);
    

    //m = 30; % Number of terms in Taylor expansion.
    int m=30;

    //% Recurrence Taylor coefficients (scaled & incl factorial terms).
    //% Flip for more accuracy in inner product calculation.
    double M = 1.0/x1;
    double* u = new double[m+1];
    double* up = new double [m+1];
    for (k=0;k<m+1;k++)
    {
        *(u+k)=0;
        *(up+k)=0;
    }
    *(u+m) = P;
    for (k=0;k<m-1;k=k+2)
    {
        *(up+m-k-1)= (k-nn*(nn+1.0)/(k+1.0))**(u+m-k)/(M*M);
        *(u+m-k-2)=*(up+m-k-1)/(k+2.0);
    }
    //% Newton iteration
    double* x1k = new double[m+1];
    for (k=0;k<m+1;k++) *(x1k+k) = 1.0;
    double step = 1.0; int l=0;double eps =  2.2204e-016;
    double sk1; double sk2;
    while ((fabs(step)>eps) && (l<10))
    {
        l++;
        sk1=0;sk2=0;
        for (k=0;k<m+1;k++)
        {
            sk1+=*(u+k)**(x1k+k);
            sk2+=*(up+k)**(x1k+k);
        }
        step = sk1/sk2/M;
        x1 = x1 - step;
        double hilfs = 1.0;
        *(x1k+m)=1.0;
        for (k=m-1; k>=0;k--)
        {
            hilfs*=(M*x1);//+*(zz+m-k-1)
            *(x1k+k) = hilfs;
        }
    }
    // Get the derivative at this root, i.e. P'(x1).
    sk2 = 0;
    for (k=0;k<m+1;k++)
    {
        sk2+=*(up+k)**(x1k+k);
    }
    double d1 = M*sk2;
    root = x1;
    ders = d1;

    delete [] u,up,x1k;
}
    
void alg1_leg(double *result_roots,double *result_ders,int mrowsx)
{
const double PI = 4* atan(1.0);
//*******xx =roots; yy = ders;
// n = length(roots);
// if mod(n,2), N = (n-1)/2; s = 1; else N = n/2; s = 0; end 

int n = mrowsx;
double nn = n;
int N; int s;
if (n%2==1)
{
   N = (n-1)/2;
   s = 1;
}
else
{
   N = n/2;
   s = 0;
} 
//% Approximate roots via asymptotic formula.
double theta;
int count = (n+4-s)/2-1;
double n3 = nn*nn*nn; double n4= n3*nn;         
double help = (nn-1.0)/(8.0*n3);
for (int k=(n-2+s)/2; k>0; k--)
{
     theta = PI* (4.0*k-1.0)/(4.0*n+2.0);
     *(result_roots+count) = (1-help-1/(384.0*n4)*(39-28/(sin(theta)*sin(theta))))*cos(theta);
     count++;     
}
double x = *(result_roots+N);

//% Number of terms in Taylor expansion.
int m=30;

//% Storage
double * hh1,*u,*up,*hh;
hh1 = new double[m+1];
u = new double[m+1];
up = new double[m+1];
hh = new double[m+1];
for (int l=0;l<m+1;l++)
{
   hh1[l] = 1.0;
   u[l]=up[l]=0.0;
}

//% Loop over all the roots we want to find (using symmetry).
for (int j=N+1;j<n;j++)
{
//% Distance to initial approx for next root (from asymptotic foruma).
   double h = *(result_roots+j)-x;
//    % Recurrence Taylor coefficients (scaled & incl factorial terms)
   double M = 1.0/h;
   double c1 = 2.0*x*h;
   double c2 = 1.0/(1.0-x*x);

  /*% Note, terms are flipped for more accuracy in inner product calculation.
    u([m+1 m]) = [0 ders(j)/M];  up(m+1) = u(m);
    for k = 0:m-2
        up(m-k) = (c1*(k+1)*u(m-k)+(k-n*(n+1)/(k+1))*u(m-k+1)/M^2)*c2;
        u(m-(k+1)) = up(m-k)/(k+2);
    end
    up(1) = 0;
  */
   *(u+m)=0; *(u+m-1)=*(result_ders+j-1)/M; *(up+m)=*(u+m-1);
   for (int k=0;k<m-1;k++)
   {
      *(up+m-k-1)= (c1*(k+1.0)**(u+m-k-1)+(k-nn*(nn+1.0)/(k+1.0))**(u+m-k)/(M*M))*c2;
      *(u+m-k-2) = *(up +m-k-1)/(k+2.0);      
   }   
   *(up)=0;
    

//  % Newton iteration
   double step = 1.0; double eps =  2.2204e-016 ; int l = 0;
   for (int k=0; k<m+1;k++)
   {
      *(hh+k)=*(hh1+k);
   }
   double sk1 , sk2;
   while ((fabs(step)>eps)&&(l<10))
   {
      l++;
      sk1 = sk2=0;
      for (int k=0; k<m+1;k++)
      {
          sk1+=*(u+k)**(hh+k);
          sk2+=*(up+k)**(hh+k);
      }
      step =  sk1/sk2/M;
      h = h -step;
      *(hh+m) = 1.0; double hilfs = 1.0;
      for (int k=m-1; k>=0;k--)
      {
         hilfs*=(M*h);//+*(zz+m-k-1)
         *(hh+k) = hilfs;
      }
   }   
//    % Update
   x = x +h;
   *(result_roots+j) = x;
   sk2 = 0;
   for (int k=0;k<m+1;k++)
   {
        sk2+=*(up+k)**(hh+k);
   }
   *(result_ders+j) = M*sk2;   
}

//% Nodes are symmetric.
for  (int k=0; k<N+s;k++)
{
   *(result_roots+k)= - *(result_roots+n-k-1);  
   *(result_ders+k)= - *(result_ders+n-k-1);
} 

delete[] hh1,hh,u,up;
}

/* the gateway function */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  double *yy,*result_roots,*result_ders;
  double  *xx;
 // mwSize mrowsy,ncolsy,mrowsx,ncolsx;
  int n; //dim
  /*  check for proper number of arguments */
  if(nrhs!=1) 
    mexErrMsgTxt("One input required.");
  if(nlhs!=2) 
    mexErrMsgTxt("Two output required.");
  
  /*  create a pointer to the input matrix x  */
  n = mxGetScalar(prhs[0]);
 
  
  /*  set the output pointer to the output matrix */
  plhs[0] = mxCreateDoubleMatrix(n,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(n,1, mxREAL);
  
  
  /*  create a C++ pointer to a copy of the output matrix */
  result_roots = mxGetPr(plhs[0]);
  result_ders = mxGetPr(plhs[1]);

    //DRIVER for alg1_leg
   for (int i=0;i<n;i++)
   {
      result_roots[i]=0;
      result_ders[i]=0;
   }
  double Pm2 = 0; double Pm1 = 1; double P;
for (int k= 0;k<n;k++)
{
    P = -k* Pm2/(k+1.0); Pm2 = Pm1; Pm1 = P;
}
if (n%2 == 1)
{
   *(result_roots+(n-1)/2-1)=0;
   *(result_ders + (n+1)/2-1)= n*Pm2;   
}   
else
{

      //even, find first root:
      double root,ders;
      alg2_Leg(P,n,root,ders);
      *(result_roots+n/2)=root;
      *(result_ders+n/2)=ders;
}
  
  /*  call the C++ subroutine */
  alg1_leg(result_roots,result_ders,n); 
}

